/*
 * _beep_brrr_01.xc
 *
 *  Created on: 2. apr. 2021
 *      Author: teig
 */


#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <i2c.h>
    #include <i2s.h>
    #include <dsp.h>
    #include <dsp_fft.h>
    #include <dsp_math.h>
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "param.h"
    #include "i2c_client_task.h"
    #include "display_ssd1306.h"
    #include "core_graphics_adafruit_gfx.h"
    #include "core_graphics_font5x8.h"
    #include "_texts_and_constants.h"
    #include "button_press.h"
    #include "pwm_softblinker.h"
    #include "maths.h"
    #include "mic_array.h"
    #include "maths_fix.h" // Needs stdint.h
    #include "mics_in_headset_out.h"
    #include "port_tile0_task.h"
    #include "_Beep_dsp_params.h"
    #include "_Beep_dsp_handling.h"
    #include "_Beep_BRRR_01.h"

    #define DO_DEBUG_PRINT         1 // 0 has consequences if (USE_XSCOPE == 1)
    #define DO_DEBUG_PRINT_SPECIAL 0 // 0 has consequences if (USE_XSCOPE == 1)
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif

typedef enum display_screen_name_t {
    SCREEN_INIT,
    // SCREEN_MIC,
    // SCREEN_MICS_CONFIG,
    SCREEN_HEADSET_GAIN_DB,
    SCREEN_ABOUT, // LAST SEEN
    SCREEN_DARK   // LAST (NOT SEEN)
} display_screen_name_t;

typedef enum {is_on, is_off} display_state_e;

#define DISPLAY_LOG_PACKETS_NUM 3 // 2 or larger

#define DO_DISPLAY_INFO_TYPE 3 // 0 "standard"
                               // 1 v0421
                               // 2 v0602 follow_mic_sample_min_max
                               // 3 v0611 spectrum max and index

typedef struct {
    display_state_e       state;
    display_screen_name_t display_screen_name;
    display_screen_name_t display_screen_name_when_into_dark;
    char                  display_ts1_chars [SSD1306_TS1_DISPLAY_VISIBLE_CHAR_LEN]; // 84 chars for display needs 85 char buffer (with NUL) when sprintf is use (use SSD1306_TS1_DISPLAY_ALL_CHAR_LEN for full flexibility)
    int                   sprintf_numchars;
    unsigned              screen_timeouts_since_last_button_countdown; // From NUM_TIMEOUTS_BEFORE_SCREEN_DARK to zero for SCREEN_DARK
    unsigned              cnt;
    unsigned              num_lost_overflowed;
    unsigned              num_samples_arr           [DISPLAY_LOG_PACKETS_NUM];
    unsigned              chan_mutual_packet_dim_arr[DISPLAY_LOG_PACKETS_NUM];
    mic_sample_t          mic_samples_arr           [DISPLAY_LOG_PACKETS_NUM];
    mic_sample_t          mic_sample_last;
    mic_sample_t          mic_sample_max; // debug IOF_MIC_0
    mic_sample_t          mic_sample_min; // debug IOF_MIC_0
    data_ch_results_t     data_ch_results;
} display_context_t;

typedef struct {
    bool     now;
    unsigned timeouts_after_last_button_cnt;
    unsigned volume_step_factor; // increasing for every repeat taken, so that going to VOLUME_MIN_DB does not feel "endless"
} repeat_t;

typedef enum {was_none, was_button, was_timeout} last_action_e;

typedef struct {
    bool     valid;
    unsigned index; // [0..BUTTONS_NUM_CLIENTS]
} button_t;

typedef struct {
    button_t         button;
    bool             pressed_ever;
    bool             button_action_taken;
    button_action_t  button_action [BUTTONS_NUM_CLIENTS];
    repeat_t         repeat;
    last_action_e    last_action; // AMUX=005
    bool             ignore_left_button_release_no_wake_from_dark; // AMUX=006 new. Since I started with LEFT_BUTTON take on released
    mic_result_t     mic_result;
    bool             send_clone_to_dsp;
    ch_data_ack_ui_t ch_data_ack_ui;
} buttons_context_t;

// For set_one_percent_ms and set_sofblink_percentages
#define SOFTBLINK_RESTARTED_ONE_PERCENT_MS        1 //  1 ms goes to 100 in 0.1 seconds -> 5 blinks per second
#define SOFTBLINK_RESTARTED_UNIT_MAX_PERCENTAGE 100
#define SOFTBLINK_RESTARTED_UNIT_MIN_PERCENTAGE   0
//
#define SOFTBLINK_DARK_DISPLAY_ONE_PERCENT_MS    30 // 30 ms goes to 100 in 3.0 seconds
#define SOFTBLINK_DARK_DISPLAY_MAX_PERCENTAGE    40
#define SOFTBLINK_DARK_DISPLAY_MIN_PERCENTAGE    10
//
#define SOFTBLINK_LIT_DISPLAY_ONE_PERCENT_MS     10 // 10 ms goes to 100 in 1.0 seconds
#define SOFTBLINK_LIT_DISPLAY_MAX_PERCENTAGE     80
#define SOFTBLINK_LIT_DISPLAY_MIN_PERCENTAGE     10

#define NUM_TIMEOUTS_PER_SECOND     2
#define NUM_TIMEOUTS_BEFORE_REPEAT (1.5 * NUM_TIMEOUTS_PER_SECOND)  // 1,5 seconds
//      BUTTON_ACTION_PRESSED_FOR_LONG_TIMEOUT_MS must be so long that it does not interfere here

// From makefile
#if (DISPLAY_FAST_DARK==1)
    #if (WARNINGS==1)
        #warning Not standard display to dark timeout
    #endif
    #define NUM_TIMEOUTS_BEFORE_SCREEN_DARK (10 * NUM_TIMEOUTS_PER_SECOND)     // 10 seconds, for screen_timeouts_since_last_button
#elif (DISPLAY_FAST_DARK==0)
    #define NUM_TIMEOUTS_BEFORE_SCREEN_DARK ((5*60) * NUM_TIMEOUTS_PER_SECOND) // 5 minutes, for screen_timeouts_since_last_button
#endif

void do_display_params_zero (display_context_t &display_context) {

    Clear_All_Pixels_In_Buffer();

    for (int index_of_char = 0; index_of_char < NUM_ELEMENTS(display_context.display_ts1_chars); index_of_char++) {
        display_context.display_ts1_chars [index_of_char] = ' ';
    }

    setTextColor(WHITE);
    setCursor(0,0);
    setTextSize(1); // SSD1306_TS1_LINE_CHAR_NUM gives 21 chars per line (but crlf takes two, if used)
}

char button_to_char (const button_t button) {
    char char_ret;

    if (button.valid) {
        char_ret = 'A' + button.index;
    } else {
        char_ret = '?';
    }
    return char_ret;
}

// MUST NOT MODIFY ANY STATE VALUES!
// 0104 kills real-time (see scope picture)
bool // i2c_ok
    Display_screen (
        display_context_t              &display_context,
        buttons_context_t              buttons_context,      // No & = no modification
        headset_gain_context_t         headset_gain_context, // No & = no modification
        client i2c_general_commands_if if_i2c_general_commands) {

    bool i2c_ok = true;
    bool do_display_print = false;

    do_display_params_zero (display_context);

    if (display_context.state == is_on) {
        switch (display_context.display_screen_name) {
            case SCREEN_INIT: {
                debug_print("%s\n", "Display_screen");
                setTextSize(2);
                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                        "Beep-BRRR");
                display_context.cnt++;
                do_display_print = true;
            } break;

            case SCREEN_HEADSET_GAIN_DB: {
                const char char_up_arrow_str[] = CHAR_UP_ARROW_STR;
                setTextSize(2);

                // See ...Beep_BRRR v0240 display KnockCome measurements.jpg display photos: SPEED_THROUGHPUT_TAG

                #if (DO_DISPLAY_INFO_TYPE == 0)
                    if (headset_gain_context.headset_gain_is_off) {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "AV -- dB\nOvfl: %u\n#: %u %u %u\nD: %u %u %u",
                            display_context.num_lost_overflowed,
                            display_context.num_samples_arr[0],
                            display_context.num_samples_arr[1],
                            display_context.num_samples_arr[2],
                            display_context.chan_mutual_packet_dim_arr[0],
                            display_context.chan_mutual_packet_dim_arr[1],
                            display_context.chan_mutual_packet_dim_arr[2]
                        );
                    } else {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "UT %d dB%s\nOvfl: %u\n#: %u %u %u\nD: %u %u %u",
                                headset_gain_context.amplitude_db_table[headset_gain_context.iof_gain],
                                (headset_gain_context.iof_gain==IOF_HEADSET_GAIN_DB_MAX) ? char_up_arrow_str : "",
                                display_context.num_lost_overflowed,
                                display_context.num_samples_arr[0],
                                display_context.num_samples_arr[1],
                                display_context.num_samples_arr[2],
                                display_context.chan_mutual_packet_dim_arr[0],
                                display_context.chan_mutual_packet_dim_arr[1],
                                display_context.chan_mutual_packet_dim_arr[2]
                            );
                    }
                #elif (DO_DISPLAY_INFO_TYPE == 1)
                    if (headset_gain_context.headset_gain_is_off) {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "AV -- dB\n%08x %08x\n%08x %08x-\n#:%u D:%u",
                            display_context.mic_samples_arr[0],
                            display_context.mic_samples_arr[1],
                            display_context.mic_samples_arr[2],
                            display_context.mic_sample_last,
                            display_context.num_samples_arr[2],
                            display_context.chan_mutual_packet_dim_arr[2]
                        );
                    } else {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "UT %d dB%s\n%08x %08x\n%08x %08x-\n#:%u D:%u",
                                headset_gain_context.amplitude_db_table[headset_gain_context.iof_gain],
                                (headset_gain_context.iof_gain==IOF_HEADSET_GAIN_DB_MAX) ? char_up_arrow_str : "",
                                display_context.mic_samples_arr[0],
                                display_context.mic_samples_arr[1],
                                display_context.mic_samples_arr[2],
                                display_context.mic_sample_last,
                                display_context.num_samples_arr[2],
                                display_context.chan_mutual_packet_dim_arr[2]
                            );
                    }
                #elif (DO_DISPLAY_INFO_TYPE == 2)
                    const signed mic_sample_min_dB = int32_dB (display_context.mic_sample_min);
                    const signed mic_sample_max_dB = int32_dB (display_context.mic_sample_max);

                    if (headset_gain_context.headset_gain_is_off) {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "AV -- dB\n %3d %08X\n%3d %08X",
                            mic_sample_max_dB,
                            display_context.mic_sample_max,
                            mic_sample_min_dB,
                            display_context.mic_sample_min
                        );
                    } else {
                        display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                            "UT %d dB%s\n %3d %08X\n%3d %08X",
                                headset_gain_context.amplitude_db_table[headset_gain_context.iof_gain],
                                (headset_gain_context.iof_gain==IOF_HEADSET_GAIN_DB_MAX) ? char_up_arrow_str : "",
                                mic_sample_max_dB,
                                display_context.mic_sample_max,
                                mic_sample_min_dB,
                                display_context.mic_sample_min
                            );
                    }
                    #elif (DO_DISPLAY_INFO_TYPE == 3)

                        char q24_squared_str [Q_FORMAT_STR_LEN];

                        q24_signed_to_str (q24_squared_str, display_context.data_ch_results.freq_max_value.val_at_ix);

                        if (headset_gain_context.headset_gain_is_off) {
                            display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                                "AV -- dB\nAT %u\n%s",
                                display_context.data_ch_results.freq_max_value.ix,
                                q24_squared_str
                            );
                        } else {
                            display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                                "UT %d dB%s\nAT %u\n%s",
                                    headset_gain_context.amplitude_db_table[headset_gain_context.iof_gain],
                                    (headset_gain_context.iof_gain==IOF_HEADSET_GAIN_DB_MAX) ? char_up_arrow_str : "",
                                    display_context.data_ch_results.freq_max_value.ix,
                                    q24_squared_str
                                );
                        }
                #elif
                    #error DO_DISPLAY_INFO_TYPE illegal value
                #endif

                do_display_print = true;
            } break;

            case SCREEN_ABOUT: {
                const char char_OE_str[]          = CHAR_OE_STR; // 
                const char char_right_arrow_str[] = CHAR_RIGHT_ARROW_STR;
                int16_t cursor_x;
                int16_t cursor_y;

                setTextSize(2);
                display_print ("Beep-BRRR\n", 10);

                cursor_x = getCursor_x();
                cursor_y = getCursor_y();
                setCursor (cursor_x, cursor_y+3);

                setTextSize(1);
                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                                      "MicArray board XMOS\nXC CODE %s\nV:%s  xT:%u.%u\n",
                                      BEEP_BRRR_DATE,    BEEP_BRRR_VERSION_STR,
                                      XCC_VERSION_MAJOR, XCC_VERSION_MINOR);
                display_print (display_context.display_ts1_chars, display_context.sprintf_numchars);

                cursor_x = getCursor_x();
                cursor_y = getCursor_y();
                setCursor (cursor_x, cursor_y+3);

                display_context.sprintf_numchars = sprintf (display_context.display_ts1_chars,
                                      "%syvind Teig",
                                      char_OE_str);
                   //                                            Beep-BRRR
                   //
                   //                                            MicArray XMOS XA
                   //                                            XC CODE JUN 18 2020
                   //                                            V:0.1.75  xT:1404.1
                   //                                            yvind Teig
                do_display_print = true;
            } break;

            default: {} break;
        }
    } else { // is_off
        do_display_print = true; // To make it dark
    }

    if (do_display_print) {
        display_print (display_context.display_ts1_chars, display_context.sprintf_numchars); // num chars not including NUL
        i2c_ok = writeToDisplay_i2c_all_buffer (if_i2c_general_commands);
        // debug_print ("%s\n", i2c_ok ? "ok2" : "err2");
    } else {}

    return i2c_ok;
} // Display_screen

#if (DO_DISPLAY_INFO_TYPE == 2)
    void restart_mic_sample_min_max (display_context_t &display_context)
    {
        display_context.mic_sample_max = MIC_ARRAY_WORD_MIN;
        display_context.mic_sample_min = MIC_ARRAY_WORD_MAX;
    }

    void follow_mic_sample_min_max (
            display_context_t  &display_context,
            const mic_sample_t mic_sample)
    {
        display_context.mic_sample_max = max (display_context.mic_sample_max, mic_sample);
        display_context.mic_sample_min = min (display_context.mic_sample_min, mic_sample);
    }
#else
    #define restart_mic_sample_min_max(x)
    #define follow_mic_sample_min_max(x,y)
#endif

void display_context_init_debug (display_context_t &display_context) {

    for (unsigned ix=0; ix<DISPLAY_LOG_PACKETS_NUM; ix++) {
        display_context.num_samples_arr           [ix] = 0;
        display_context.chan_mutual_packet_dim_arr[ix] = 0;
        display_context.mic_samples_arr           [ix] = 0;
    }

    display_context.mic_sample_last = 0;
    restart_mic_sample_min_max (display_context);
}


void display_context_init (display_context_t &display_context) {

    display_context.display_screen_name                         = SCREEN_ABOUT; // Was SCREEN_INIT before 0250
    display_context.state                                       = is_on;
    display_context.screen_timeouts_since_last_button_countdown = NUM_TIMEOUTS_BEFORE_SCREEN_DARK;
    display_context.cnt                                         = 0;
    display_context.num_lost_overflowed                         = 0;
    display_context.data_ch_results.freq_max_value.val_at_ix    = 0;
    display_context.data_ch_results.freq_max_value.ix           = 0;

    display_context_init_debug(display_context);
}

void buttons_context_init (buttons_context_t &buttons_context) {
    buttons_context.button.valid = false;
    //
    buttons_context.send_clone_to_dsp = SEND_LONG_MESSAGE_ACROSS_TILES_INIT;
    buttons_context.ch_data_ack_ui.any_button_any_event.cnt = 0;
}

void menu_result_init (menu_result_t &menu_result) {
    menu_result.headset_gain_updated = true;
    menu_result.headset_gain         = HEADSET_GAIN_DEFAULT;
    menu_result.headset_gain_is_off  = HEADSET_GAIN_IS_OFF_VALUE_DEFAULT;
}

// [[combinable]]
void Handle_Beep_BRRRR_task_b (
         server button_if_1             if_button [BUTTONS_NUM_CLIENTS],
         out port                       p_leds_00_07,
         client softblinker_if          if_softblinker,
         client i2c_general_commands_if if_i2c_general_commands,
         out port                       p_display_notReset,
         out port                       p_scope_1d_1n,
         streaming chanend              sch_ab,
         chanend                        ch_ba,
         chanend                        ch_bx,
         streaming chanend              sch_yb)
{
    #if (USE_XSCOPE == 1)
        delay_milliseconds (DELAY_FIRST_PRINT_MS);
    #endif
    debug_print("%s %s Handle_Beep_BRRRR_task_b started\n\tProcess @ %u Hz with %u samples per batch\n\tFilters A:%u Y:%u\n\tXTC %u.%u\n\tDisplay: %u * %u\n",
            BEEP_BRRR_DATE,
            BEEP_BRRR_VERSION_STR,
            PROCESS_FREQUENCY_HZ,
            NUM_SAMPLES_PER_BATCH,
            NUM_SECTIONS_FILTER_A, NUM_SECTIONS_FILTER_Y,
            XCC_VERSION_MAJOR,     XCC_VERSION_MINOR,
            SSD1306_LCDWIDTH,      SSD1306_LCDHEIGHT);

    q24_test_conversion_iff(); // Provided DO_Q24_TEST_CONVERSION==1 in "maths_fix.h"
    q24_test_math_iff();       // Provided DO_Q24_TEST_MATH==1       in "maths_fix.h"

    unsigned one_percent_ms = SOFTBLINK_DARK_DISPLAY_ONE_PERCENT_MS;
    unsigned cnt = 0;
    timer    tmr;
    time32_t time_ticks;

    display_context_t      display_context;
    buttons_context_t      buttons_context;
    headset_gain_context_t headset_gain_context;
    //
    data_sch_phase1_t phase1_sch_mic_sample;
    data_ch_phase2_t  phase2_ch_ack_etc;
    data_ch_phase3_t  phase3_ch_mic_samples;
    data_sch_knock_t  data_sch_knock;

    #if (DEBUG_MIC_SAMPLES_INCREASING > 0)
        mic_sample_t debug_sliding_mic_sample = COUNTING_MIC_SAMPLE_INIT;
    #endif

    menu_result_t menu_result;

    set_headset_gain (do_headset_gain_default, headset_gain_context);
    menu_result_init (menu_result);

    // STARTUP
    if_softblinker.set_one_percent_ms (one_percent_ms);
    if_softblinker.set_sofblink_percentages (SOFTBLINK_RESTARTED_UNIT_MAX_PERCENTAGE, SOFTBLINK_RESTARTED_UNIT_MIN_PERCENTAGE);

    buttons_context_init (buttons_context);
    display_context_init (display_context);

    Adafruit_GFX_constructor   (SSD1306_LCDWIDTH, SSD1306_LCDHEIGHT);
    Adafruit_SSD1306_i2c_begin (if_i2c_general_commands, p_display_notReset);

    #if (DEBUG_SET_PIXEL_IN_BUFFER_TEST==1)
        setPixel_in_buffer_test ();
        writeToDisplay_i2c_all_buffer (if_i2c_general_commands);
        delay_milliseconds (1000);
        xassert(false);
    #endif

    DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_3);
    Display_screen         (display_context, buttons_context, headset_gain_context, if_i2c_general_commands);
    DEBUG_PORTS_1D_1N_LOW  (p_scope_1d_1n,PORTS_1D_1N_AT_3);

    delay_milliseconds (2000); // To be able to read the display

    DEBUG_1PORT_LOW (p_scope_1d_1n);

    { // To avoid unnecassary num_lost_overflowed at start
        data_ch_init_synch_t synch = 1; // Value not significant
        ch_ba <: synch;
    }

    tmr :> time_ticks; // Immediately;

    while(1) {
        [[ordered]] // Ok, but not strictly necessary
        select {
            case sch_yb :> data_sch_knock : { // Atomic and deadlock free provided channel changes direction now:
                DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_2);
                // Atomic deadlock-free exchange of data
                if (data_sch_knock.knock_reason == out_dsp_result_after_ack) { // out there is in here:

                    data_sch_ack_t data_sch_ack;
                    buttons_context.ch_data_ack_ui.data_sch_ack.ack_reason = out_dsp_result;

                    sch_yb <: buttons_context.ch_data_ack_ui; // channel changed direction
                    sch_yb :> display_context.data_ch_results;

                } else if (data_sch_knock.knock_reason == in_ui_in_ack) { // out there is in here:

                    data_sch_ack_t data_sch_ack;
                    buttons_context.ch_data_ack_ui.data_sch_ack.ack_reason = in_ui;

                    sch_yb <: buttons_context.ch_data_ack_ui; // channel changed direction

                } else {
                    xassert (false);
                }

                display_context.display_screen_name = SCREEN_HEADSET_GAIN_DB;

                #if (DO_DISPLAY_SCREEN_OFTEN==1)
                   // v0629
                    // BLIPS here, but not every time!
                    // If I move Display_screen to timerafter, the BLIP moves with it
                    // headset_gain_context.headset_gain_is_off not modified here
                    DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_3);
                    Display_screen         (display_context, buttons_context, headset_gain_context, if_i2c_general_commands);
                    DEBUG_PORTS_1D_1N_LOW  (p_scope_1d_1n,PORTS_1D_1N_AT_3);
                #else
                    // delay_milliseconds(30); // ONLY during testing. TODO remove
                #endif

                DEBUG_PORTS_1D_1N_LOW (p_scope_1d_1n,PORTS_1D_1N_AT_2);
            } break;

            case sch_ab :> phase1_sch_mic_sample : {

                phase2_ch_ack_etc.menu_result = menu_result;

                if (phase1_sch_mic_sample.chan_mutual_packet_dim == NUM_000) {

                    phase2_ch_ack_etc.chan_mutual_packet_dim = NUM_000; // For the ack

                    follow_mic_sample_min_max (
                            display_context,
                            phase1_sch_mic_sample.iff.valid.mic_sample);

                    ch_ba <: phase2_ch_ack_etc;

                    send_clone_iff (
                            buttons_context.send_clone_to_dsp,
                            phase1_sch_mic_sample,
                            ch_bx); // Observe ch_bx sent to again, below

                    VERIFY_MIC_SAMPLES_DELTA (20, 1, debug_sliding_mic_sample,  phase1_sch_mic_sample.iff.valid.mic_sample, NUM_000);

                } else { // >= NUM_001

                    DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_1);
                    phase2_ch_ack_etc.chan_mutual_packet_dim = phase1_sch_mic_sample.chan_mutual_packet_dim; // For round trip

                    // === ATOMIC TRANSACTION (BEGIN) ===
                    // SEND ACK:
                    ch_ba <: phase2_ch_ack_etc;
                    // RECEIVE DATA
                    // WAS SENT BY send_variant_protocol_then_shift_down

                    receive_variant_protocol_send_clone_iff (
                            sch_ab,
                            phase2_ch_ack_etc.chan_mutual_packet_dim,
                            phase3_ch_mic_samples,             // New values, also sent off iff
                            buttons_context.send_clone_to_dsp, // if true match receive_variant_protocol, from buffer_task_x
                            phase1_sch_mic_sample,             // To send off iff
                            ch_bx);                            // Observe ch_bx sent to again, below

                    VERIFY_MIC_SAMPLES_DELTA (21, 1,
                            debug_sliding_mic_sample,
                            phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[0],
                            phase3_ch_mic_samples.samples.rec_max_msg.nums.num_samples);

                    const unsigned     num_samples = phase3_ch_mic_samples.samples.rec_max_msg.nums.num_samples;
                    const mic_sample_t next_sample = phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[num_samples-1];

                    for (unsigned ix=0; ix<num_samples; ix++) {
                        follow_mic_sample_min_max (
                                display_context,
                                phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[ix]);
                    }

                    VERIFY_MIC_SAMPLES_DELTA (22, num_samples-1, debug_sliding_mic_sample, next_sample, -1);

                    for (unsigned ix=0; ix<(DISPLAY_LOG_PACKETS_NUM-1); ix++) {
                        display_context.num_samples_arr           [ix] = display_context.num_samples_arr           [ix+1];
                        display_context.chan_mutual_packet_dim_arr[ix] = display_context.chan_mutual_packet_dim_arr[ix+1];
                        display_context.mic_samples_arr           [ix] = display_context.mic_samples_arr[ix+1];
                    }
                    unsigned mic_num_samples = phase3_ch_mic_samples.samples.rec_max_msg.nums.num_samples;
                    //
                    display_context.num_samples_arr           [DISPLAY_LOG_PACKETS_NUM-1] = mic_num_samples;
                    display_context.chan_mutual_packet_dim_arr[DISPLAY_LOG_PACKETS_NUM-1] = phase2_ch_ack_etc.chan_mutual_packet_dim;
                    display_context.num_lost_overflowed                                   = phase3_ch_mic_samples.samples.rec_max_msg.nums.num_lost_overflowed;
                    display_context.mic_samples_arr           [DISPLAY_LOG_PACKETS_NUM-1] = phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[0];
                    //
                    // v0421 twelve bins containing like: 10 11 12 13 14 15  16 17 18 19 20 21,then [last]-[first]= 11
                    display_context.mic_sample_last =
                            phase3_ch_mic_samples.samples.rec_max_msg.mic_samples.samples[mic_num_samples-1];

                    // === ATOMIC TRANSACTION (END) ===
                    DEBUG_PORTS_1D_1N_LOW (p_scope_1d_1n,PORTS_1D_1N_AT_1);
                }

                menu_result.headset_gain_updated = false;

            } break;

            case tmr when timerafter (time_ticks) :> void : {
                time_ticks += (XS1_TIMER_HZ/NUM_TIMEOUTS_PER_SECOND);
                buttons_context.button.valid = false;

                if (display_context.state == is_off) {
                    // No code, don't exercise the display and I2C
                } else if (display_context.screen_timeouts_since_last_button_countdown > 0) {
                    display_context.screen_timeouts_since_last_button_countdown--;
                } else { // == 0
                    display_context.state = is_off;
                    DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_3);
                    Display_screen         (display_context, buttons_context, headset_gain_context, if_i2c_general_commands);
                    DEBUG_PORTS_1D_1N_LOW  (p_scope_1d_1n,PORTS_1D_1N_AT_3);
                }

            } break;

            // If the button is "stable", react when the I/O pin changes value
            case if_button[unsigned iof_button].button (const button_action_t button_action) : {

                // 0 BUTTON_ACTION_VOID,
                // 1 BUTTON_ACTION_PRESSED,
                // 2 BUTTON_ACTION_PRESSED_FOR_LONG, // BUTTON_ACTION_PRESSED_FOR_LONG_TIMEOUT_MS
                // 3 BUTTON_ACTION_RELEASED          // Not after BUTTON_ACTION_PRESSED_FOR_LONG

                one_percent_ms = (SOFTBLINK_RESTARTED_ONE_PERCENT_MS * (iof_button+1));

                #if (USE_XSCOPE == 1)
                    // Would have destroyed all timing
                #elif (USE_XSCOPE == 0)
                    debug_print_special("A Handle_Beep_BRRRR_task_b [%u]=%u\n", iof_button, button_action);
                #endif

                if_softblinker.set_one_percent_ms (one_percent_ms);

                cnt++;

                buttons_context.ch_data_ack_ui.any_button_any_event.cnt++; // sent if send_clone_to_dsp

                if (button_action == BUTTON_ACTION_PRESSED) {
                    #if (USE_PORTS_1D_1N_AT == PORTS_1D_1N_AT_5)
                        DEBUG_PORTS_1D_1N_VAL (p_scope_1d_1n,true,PORTS_1D_1N_AT_5);
                    #endif
                    display_context.state = is_on;
                    display_context.screen_timeouts_since_last_button_countdown = NUM_TIMEOUTS_BEFORE_SCREEN_DARK;
                    buttons_context.button.valid = true;
                    buttons_context.button.index = iof_button;
                    //
                    switch (iof_button) {
                        case IOF_BUTTON_LEFT: {
                            set_headset_gain (do_headset_gain_zero, headset_gain_context);
                        } break;
                        case IOF_BUTTON_CENTER: {
                            set_headset_gain (do_headset_gain_decrease, headset_gain_context);
                        } break;
                        case IOF_BUTTON_RIGHT: {
                            set_headset_gain (do_headset_gain_increase, headset_gain_context);
                        } break;
                        case IOF_BUTTON_FAR_RIGHT: {
                            set_headset_gain (do_headset_gain_default, headset_gain_context);
                        } break;
                        // default: fault by xC runtime
                    }
                    display_context.display_screen_name = SCREEN_HEADSET_GAIN_DB;
                    //
                    menu_result.clear_debug_vars = (iof_button == IOF_BUTTON_LEFT); // Return value too
                    //
                    if (menu_result.clear_debug_vars) {
                        buttons_context_init (buttons_context);
                        display_context_init_debug (display_context);
                    } else {}

                    menu_result.headset_gain         = headset_gain_context.gain;
                    menu_result.headset_gain_is_off  = headset_gain_context.headset_gain_is_off;
                    menu_result.headset_gain_updated = true;

                    DEBUG_PORTS_1D_1N_HIGH (p_scope_1d_1n,PORTS_1D_1N_AT_3);
                    Display_screen         (display_context, buttons_context, headset_gain_context, if_i2c_general_commands);
                    DEBUG_PORTS_1D_1N_LOW  (p_scope_1d_1n,PORTS_1D_1N_AT_3);

                    restart_mic_sample_min_max (display_context);

                    phase2_ch_ack_etc.menu_result = menu_result;
                } else if (button_action == BUTTON_ACTION_RELEASED) {
                    #if (USE_PORTS_1D_1N_AT == PORTS_1D_1N_AT_5)
                        DEBUG_PORTS_1D_1N_VAL (p_scope_1d_1n,false,PORTS_1D_1N_AT_5);
                    #endif
                } else if (button_action == BUTTON_ACTION_PRESSED_FOR_LONG) {
                    // BUTTON_ACTION_RELEASED will not follow, so treat this as BUTTON_ACTION_RELEASED:
                    #if (USE_PORTS_1D_1N_AT == PORTS_1D_1N_AT_5)
                        DEBUG_PORTS_1D_1N_VAL (p_scope_1d_1n,false,PORTS_1D_1N_AT_5);
                    #endif
                } else {
                    // BUTTON_ACTION_VOID
                    // No code
                }
            } break;
        }
    }
};
