/*
 * _Beep_dsp_params.h
 *
 *  Created on: 10. mai 2022
 *      Author: oyvindteig
 *      About:  This file does not have any corersponding .xc file
 */


#ifndef BEEP_DSP_PARAMS_H_
    #define BEEP_DSP_PARAMS_H_

    // ==========
    // FILTERING
    //
    // XMOS
    //     Filter coefficients are ordered as [b0,b1,b2,-a1,-a2].
    // https://www.dsprelated.com/showarticle/1137.php
    //     Biquad (second-order) lowpass all-pole filter
    // https://www.dsprelated.com/thread/5327/how-to-obtain-biquad-coefficients
    //     For a biquad, the feedback coeffs are [1 a1 a2] and the
    //     feedforward coeffs are [b0 b1 b2].
    //     The a0 coefficient does not appear explicitly in the filter structure
    // -> https://en.wikipedia.org/wiki/Digital_biquad_filter
    // -> http://shepazu.github.io/Audio-EQ-Cookbook/audio-eq-cookbook.html

    // See example in app_design.xc in AN00209_app_design
    // dsp_design.h in lib_dsp

    typedef struct {
        int64_t padding1_for_dw_aligned;
        int32_t coeffs [DSP_NUM_COEFFS_PER_BIQUAD];
        //
        int64_t padding2_for_dw_aligned;
        int32_t states [DSP_NUM_STATES_PER_BIQUAD];
    } iir_biquad_params_t;

    // From dsp_filters.h
    //     DSP_NUM_COEFFS_PER_BIQUAD = 5
    //     DSP_NUM_STATES_PER_BIQUAD = 4

    #define NUM_SECTIONS_FILTER_A 1 // 0,1,2 or larger, before downsampling (IIR_CASCADE_DEPTH)
    //
    #if ((PROCESS_FREQUENCY_HZ == 16000) and (SAMPLING_FREQUENCY_HZ == 48000))
        // Handled by mics_in_headset_out_task_a in mics_in_headset_out.xc
        // 16 KHz, but Nyqusit is 8 kHz. 8/48 = 0.167
        #define FILTER_A_FREQ_MINUS_3DB_NORMALIZED 0.167 // 0.0 <= F < 0.5.
        #define FILTER_A_Q_FACTOR                  0.707 // 1/sqrt(2) = f0/bandwidth
    #else
        #error
    #endif

    #define NUM_SECTIONS_FILTER_Y 2 // 0,1,2 or larger before FFT (IIR_CASCADE_DEPTH)
    //
    #if ((PROCESS_FREQUENCY_HZ == 16000) and (NUM_MIC_SAMPLES == 512)) // NUM_SAMPLES_PER_BATCH
        // Handled by dsp_task_y in _Beep_dsp_handling.xc
        // 512 samples and split after the FFT -> 4 kHz, which is also one fourth
        // 16 kHz and need to go to 4 kHz. Don't know if Nyquist is here. Se 4/16 = 0.25
        #define FILTER_Y_FREQ_MINUS_3DB_NORMALIZED 0.25  // 0.0 <= F < 0.5.
        #define FILTER_Y_Q_FACTOR                  0.707 // 1/sqrt(2) = f0/bandwidth
    #elif ((PROCESS_FREQUENCY_HZ == 16000) and (NUM_MIC_SAMPLES == 1024)) // NUM_SAMPLES_PER_BATCH
        // Handled by dsp_task_y in _Beep_dsp_handling.xc
        // 1024 samples and split after the FFT -> 4 kHz, which is also one fourth
        // 16 kHz and need to go to 4 kHz. Don't know if Nyquist is here. Se 4/16 = 0.25
        #define FILTER_Y_FREQ_MINUS_3DB_NORMALIZED 0.25  // 0.0 <= F < 0.5.
        #define FILTER_Y_Q_FACTOR                  0.707 // 1/sqrt(2) = f0/bandwidth
    #else
        #error
    #endif

    #define FILTER_Q_FORMAT 28 // 28 1 sign bit, 3 bits integer, 28 bits fractionals

#else
    #error NESTED .h FILE
    // Needs:
    // mics_in_headset_out.h
    // dsp.h (and then comes dsp_filters.h) in lib_dsp

#endif /* BEEP_DSP_PARAMS_H_ */
