/*
 * maths.xc
 *
 *  Created on: 7. sep. 2021
 *      Author: oyvindteig
 */

#define INCLUDES
#ifdef INCLUDES
    #include <platform.h>
    #include <xs1.h>
    #include <iso646.h> // not etc.
    #include <stdint.h> // int32_t etc
    #include <math.h>

    #include "maths.h"
#endif

// See https://en.wikipedia.org/wiki/Decibel#Acoustics_2
// float, double and long double are all 32 bits here! See
//     https://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#data_types_long_long_float_double_and_long_double

// ARGUMENT POSITIVE -> POSITIVE dB
// ARGUMENT NEGATIVE -> NEGATIVE dB (by math values between 1 and 0 yield negative dB and negative yield NaN)
// ARGUMENT ZERO     -> ZERO     dB (by math Inf)

// Q number format as in dsp_qformat.h not used

int32_t int32_dB (const int32_t arg) {

    int32_t ret_dB;

    if (arg == 0) {
        ret_dB = 0; // (by math NaN or Inf)
    } else {
        float dB;
        if (arg > 0) {
            dB =    20 * (log10 ((float)   arg ));
        } else { // if (arg < 0) {
            dB = - (20 * (log10 ((float) -(arg) )));
        }
        ret_dB = (int32_t) dB;
    }

    return ret_dB;
}

int16_t int16_dB (const int16_t arg) {

    return (int16_t) int32_dB ((int32_t) arg);
}
