// Much from "Copyright (c) 2015-2016, XMOS Ltd, All rights reserved"
//
// ======================================================
// See "mic_array_conf.h" in lib_mic_array user guide PDF
// ======================================================
// Teig: Taken from AN00220

#ifndef MIC_ARRAY_CONF_H_
    #define MIC_ARRAY_CONF_H_

    // =============================
    // FOR lib_mic_array IN:
    //    decimate_to_pcm_4ch.S asm
    //      mic_array_frame.h
    //        pdm_rx.S asm
    //
    #define MIC_ARRAY_MAX_FRAME_SIZE_LOG2 0 // Defines MIC_ARRAY_NUM_SAMPLES_PER_MIC in audio.data
    //
    #define MIC_ARRAY_NUM_MICS_LOG2       3 // 2exp3 = 8 = MIC_ARRAY_NUM_MICS
    #define MIC_ARRAY_NUM_MICS                         8 // Must have "8 mics" (=7) since directive

    //       # MIC_ARRAY_MAX_FRAME_SIZE_LOG2
    // 2 exp(0)  =    1 MIC_ARRAY_NUM_SAMPLES_PER_MIC sample  per microphone in audio.data in example (AN00219)
    // 2 exp(1)  =    2 MIC_ARRAY_NUM_SAMPLES_PER_MIC samples per microphone in audio.data in example
    // 2 exp(2)  =    4 MIC_ARRAY_NUM_SAMPLES_PER_MIC samples per microphone in audio.data in example
    // 2 exp(3)  =    8 MIC_ARRAY_NUM_SAMPLES_PER_MIC samples per microphone in audio.data in example (AN00220)
    // ...
    // 2 exp(13) = 8192 MIC_ARRAY_NUM_SAMPLES_PER_MIC samples per microphone in audio.data in example

    // The four channel decimators (mic_array_pdm_rx()) output frames of either
    // time domain audio or FFT ready audio prepared for an FFT. The define
    // MIC_ARRAY_MAX_FRAME_SIZE_LOG2 (found in mic_array_conf.h) should be used
    // to allocate the arrays to store the frames. This means that all frames
    // structures will allocate enough memory to allow for a frame size of
    // two to the power of MIC_ARRAY_MAX_FRAME_SIZE_LOG2 regardless of the size
    // used in the decimator_config_common.
    // It is recommended that the frame_size_log2 field of decimator_config_common
    // is always set to MIC_ARRAY_MAX_FRAME_SIZE_LOG2. Equally, the define MIC_ARRAY_NUM_MICS
    // is used for allocating the memory for the frame structure. This must be set to a multiple of 4.

    // The mic_array board has 7 Infineon IM69D130 MEMS microphones
    // "Count of microphones(channels) must be set to a multiple of 4", ie 4 or 8

    // If the decimation factor is changed the the coefs_array of mic_array_decimator_config_t
    // must also be changed. Done below

    // ==============================================================================
    // THE REST OF THIS FILE IS PICKED UP BY APPLICATION IF the lib_mic_array is USED
    //

    // LORES vs. HIRES
    //   Observe that AN00219_app_lores_DAS_fixed does it this "low resolution" (lores) way.
    //   There also is an alternative AN00217_app_high_resolution_delay_example "high resolution"
    //   (hires) which is available with the mic_array_hires_delay task. Also AN00218_app_hires_DAS_fixed.
    //   From lib_mic_array user guide: "Additionally, the high resolution delay task can be inserted
    //   interface and the decimators in the following fashion (page 8). Note, using
    //   between the PDM the high resolution delay consumes an extra logical core.
    //   Also this mentioned here: AN00220_app_phase_aligned_example

    // ================
    // HEADSET_OUT_PART
    // ================

    #define TEST_SINE_480HZ 0 // 0 standard input from mics
                              // 1 from sine table, 480Hz exactly only if SAMPLING_FREQUENCY_HZ is 48000
                              // Observe SINE_480HZ_DIV_FACTOR

    // Based mostly on AN00219_app_lores_DAS_fixed. See "LORES" above.

    // ========================== #####  REBUILD ALL AFTER MOD OF THESE(!?)
    //                            #####  = Clean Project, Build Project
    //                            #####
    #define PROCESS_FREQUENCY_HZ  16000 // p-freq : SAMPLING_FREQUENCY_HZ is multiple
    #define NUM_SAMPLES_PER_BATCH  1024 // p-num  : Power of 2. Default for v0829
            //
            // === SAMPLING ==  ================== FFT =================    ==== SPLITTING IN 2 ====
            // p-freq  p-num -> Nyquist-f  FFT-bins  HZ_PER_BIN batch-ms -> spectrum-ms f-max ix-max
            //  16000 /  512 ->      8000 /    256 = 31.25            32 ->          16  4000  [127]
            //  16000 / 1024 ->      8000 /    512 = 15.625           64 ->          32  4000  [255]
            //   8000 / 1024 ->      4000 /    512 =  7.8125         128 ->          64  2000  [255]
            //   8000 / 2048 ->      4000 /   1024 =  3.90625        256 ->         128  2000  [511]

    #define USE_BEAMFORMING   1 // Always associate with MIC_ARRAY_NUM_MICS==8. Meaningless with TEST_SINE_480HZ
    #define DECIMATION_FACTOR 2 // 48 kHz (decimate = downsample)

    #define MIC_ARRAY_NUM_SAMPLES_PER_MIC (1 << MIC_ARRAY_MAX_FRAME_SIZE_LOG2) // 1 (for 219), 2, 4, 8..

    #define DO_WAITS_IN_MIC_ARRAY_GET_NEXT_TIME_DOMAIN_FRAME 0 // 0: do some wait in task first
                                                               // 1: wait in mic_array_get_next_time_domain_frame

    #define NUM_COMPLEX_PARTS 2 // .re and .im (using a define to distingush between "other 2")
    #define NUM_FREQ_SPECTRA  2 // after FFT if real signal input as TWO sequences in .re and .im

    // From the PDF doc XM010267
    // 48 kHz, 24 kHz, 16 kHz, 12 kHz and 8 kHz output sample rate by default (3.072MHz PDM clock)
    // Configurable frame size from 1 sample to 8192 samples plus 50% overlapping frames option,
    // chapter 11 "Four Channel Decimator"
    //
    // 96 kHz divided by 2, 4, 6, 8 or 12. See "fir_coefs.h" and "fir_coefs.xc" in lib_mic_array

    #define COEF_ARRAYS           g_third_stage_div_2_fir  // [126]
    #define SAMPLING_FREQUENCY_HZ 48000                    // 48 kHz =DAC_OUTPUT_SAMPLE_FREQUENCY_HZ  Observe TEST_SINE_480HZ.
    #define HEADSET_FREQUENCY_HZ  SAMPLING_FREQUENCY_HZ    // A MUST! HEADSET DAC CS43L21 and PLL only seem to work at 48 kHz

    // My local decimation (decimate = downsample) is a downsampling by an integer, just picking out
    // every N'th sample. Observe that aliasing filter applies! Should I neeed to downsample my some other number
    // (like based on 44.1 kHz), I should look in the lib_src sample rate conversion library
    //
    #define PROCESS_EVERY_N_SAMPLE (SAMPLING_FREQUENCY_HZ / PROCESS_FREQUENCY_HZ) // Used in SAMPLES_BUF_DIM

    #define FIR_GAIN_COMPENSATION FIR_COMPENSATOR_DIV_2

    #if (NUM_SAMPLES_PER_BATCH == 512)
        #define NUM_MIC_SAMPLES_LOG2 9 // 2exp9 = 512 real values

        // Some macros (like FFT_SINE()) won't allow NUM_MIC_SAMPLES_LOG2 or expressions
        //
        #define NUM_MIC_SAMPLES          512
        #define NUM_FFT_POINTS_COMPLEX   256 // 512 / NUM_COMPLEX_PARTS
        #define NUM_FREQ_POINTS_COMPLEX  128 // 256 / NUM_FREQ_SPECTRA
        #define NUM_FREQ_POINTS_REAL     128 // Same as above

    #elif (NUM_SAMPLES_PER_BATCH == 1024)
        //
        //      FIRST_PART               LAST_PART
        //      ------------------------|------------------------ 1024 NUM_MIC_SAMPLES_VAL
        //
        //        (.re)     ------------------------              512 first part
        //                            TO_FFT                      NUM_FFT_POINTS_COMPLEX
        //        (.im)     ------------------------              512 left part
        //
        //             FIRST_FREQ               LAST_FREQ         FIRST_PART and LAST_PART individual spectra
        //            ------------    .re      -----------        256 + 256
        //              FROM_FFT                 FROM FFT         NUM_FREQ_POINTS_COMPLEX
        //            ------------    .im      -----------        256 + 256

        #define NUM_MIC_SAMPLES_LOG2 10 // 2exp10 = 1024 real values

        // Some macros (like FFT_SINE()) won't allow NUM_MIC_SAMPLES_LOG2 or expressions
        //
        #define NUM_MIC_SAMPLES         1024
        #define NUM_FFT_POINTS_COMPLEX   512 // 1024 / NUM_COMPLEX_PARTS
        #define NUM_FREQ_POINTS_COMPLEX  256 //  512 / NUM_FREQ_SPECTRA
        #define NUM_FREQ_POINTS_REAL     256 // Same as above

    #elif (NUM_SAMPLES_PER_BATCH == 2048)

        #define NUM_MIC_SAMPLES_LOG2 11 // 2exp11 = 2048 real values

        // Some macros (like FFT_SINE()) won't allow NUM_MIC_SAMPLES_LOG2 or expressions
        //
        #define NUM_MIC_SAMPLES         2048
        #define NUM_FFT_POINTS_COMPLEX  1024 // 2048 / NUM_COMPLEX_PARTS
        #define NUM_FREQ_POINTS_COMPLEX  512 // 1024 / NUM_FREQ_SPECTRA
        #define NUM_FREQ_POINTS_REAL     512 // Same as above

    #elif ((1<<NUM_MIC_SAMPLES_LOG2) != NUM_MIC_SAMPLES)
        #error Power of 2 required
    #else
        #error NUM_SAMPLES_PER_BATCH
    #endif

    #if ((SAMPLING_FREQUENCY_HZ % PROCESS_FREQUENCY_HZ) != 0)
        #error Multiplies required
    #endif

#else
    #error NESTED .h FILE
#endif
