/*
 * mics_in_headset_out.h
 *
 *  Created on: 22. apr. 2021
 *      Author: teig
 *
 *      Needed include files, see bottom of this file
 */

#ifndef MICS_IN_HEADSET_OUT_H_
    #define MICS_IN_HEADSET_OUT_H_

    // =============
    // MICS_IN PART
    // =============

    #ifndef MIC_ARRAY_WORD_LENGTH_SHORT
        #error MIC_ARRAY_WORD_LENGTH_SHORT not defined
        // #if not in Makefile, it's defaulted to 0 in mic_array_frame.h
    #else
        // in mic_array_frame.h in lib_mic_array
        //
        // From "mic_array_conf.h" in lib_mic_array user guide PDF
        // MIC_ARRAY_WORD_LENGTH_SHORT
        //     If this define is set to non-zero then this configures the output word length to be a 16 bit
        //     short otherwise its left as 32 bit word length output. All internal processing will be done at
        //     32 bits, only during the write to frame memory will the truncation happen.
        // This is done in mic_array_init_time_domain_frame and mic_array_get_next_time_domain_frame
        //
        #if (MIC_ARRAY_WORD_LENGTH_SHORT==0) // THIS
            typedef int32_t mic_sample_t; // Data in 16 first LSB bits, ..
            #define MIC_ARRAY_WORD_DB(arg) int32_dB(arg)
            #define MIC_ARRAY_WORD_MIN INT_MIN
            #define MIC_ARRAY_WORD_MAX INT_MAX
        #elif (MIC_ARRAY_WORD_LENGTH_SHORT==1)
            typedef int16_t mic_sample_t; // ..here these 16 bits are all that's present
            #define MIC_ARRAY_WORD_DB(arg) int16_dB(arg)
            #define MIC_ARRAY_WORD_MIN SHRT_MIN
            #define MIC_ARRAY_WORD_MAX SHRT_MAX
        #endif
    #endif

    typedef int mic_data_t; // Must be int since mic_array_decimator_config_t has int as field 2

    typedef unsigned mics_direction_t; // [NUM_CIRCULAR_MICS] = [0..(NUM_CIRCULAR_MICS-1)]

    typedef struct {
        bool             use_beamforming; // USE_BEAMFORMING == 1
        mics_direction_t mics_direction;
    } directive_mics_t;

    #if (USE_BEAMFORMING == 1) // Always associate with MIC_ARRAY_NUM_MICS==8

        typedef unsigned delay_mics_num_samples_t;

        #define MAX_DELAY_SAMPLE_CNT 16
        #define NUM_PHYSICAL_MICS    (MIC_ARRAY_NUM_MICS - 1) // 7
        #define NUM_CIRCULAR_MICS    (NUM_PHYSICAL_MICS  - 1) // 6

        // Based on the spreadsheet mic_array_das_beamformer_calcs.xls,
        // which can be found in the root directory of this app
        #define DELAY_MICS_HEAD_DEFAULT 0
        #define DELAY_SAMPLE_COUNT_OF_0                      43
        #define DELAY_SAMPLE_COUNT_ONE_METER_THIRTY_DEGREES {   0, 3, 8, 11, 8, 3 } // NUM_CIRCULAR_MICS

    #elif (USE_BEAMFORMING == 0) // Always associate with MIC_ARRAY_NUM_MICS==4

        #define NUM_PHYSICAL_MICS MIC_ARRAY_NUM_MICS

    #else
        #error USE_BEAMFORMING error
    #endif
    //
    #define START_CIRCULAR_MICS 3 // Assume this is the meaning in AN00219 in formula in set_dir

    #ifndef MIC_ARRAY_NUM_MICS
        #error mic_array_conf.h not seen
    #endif

    #define NUM_4MICS_DECIMATOR_CHANS 4 // (DECIMATOR_CH_COUNT) Simply because mic_array_decimate_to_pcm_4ch takes 4 mics
    #define NUM_4MIC_DECIMATORS       (MIC_ARRAY_NUM_MICS / NUM_4MICS_DECIMATOR_CHANS) // (DECIMATOR_COUNT) 8 channels requires 2 decimators:
    //                                // 2/1 when MIC_ARRAY_NUM_MICS == 8/4
    #define FRAME_BUFFER_COUNT        2 // AN00220/AN00219: The minimum of 2 will suffice for these examples (when DECIMATOR_NO_FRAME_OVERLAP)

    #define NUM_DATA_X (NUM_4MIC_DECIMATORS         * NUM_4MICS_DECIMATOR_CHANS) // 8/4 when MIC_ARRAY_NUM_MICS == 8/4
    #define NUM_DATA_Y (THIRD_STAGE_COEFS_PER_STAGE * DECIMATION_FACTOR)        // 192 = 32*6

    // ================
    // HEADSET_OUT PART
    // ================

    typedef sfix16_t headset_gain_t; // signed 32 bits even if never below 0, sine we're multiplying with mic_data_t

    #define _HEADSET_GAIN_UNITY_BITPOS FIX16_UNITY_BITPOS
    #define _HEADSET_GAIN_UNITY        (1 << _HEADSET_GAIN_UNITY_BITPOS)

    typedef enum do_headset_gain_e {
        do_headset_gain_zero,
        do_headset_gain_default,
        do_headset_gain_max,
        do_headset_gain_increase,
        do_headset_gain_decrease
    } do_headset_gain_e;

    // Observe that AN00219 and mic_array_conf.h MIC_ARRAY_FIXED_GAIN work with shifting
    // bits (ie. div/mul by 2,4,8,16 etc), not a dB scale as I have chosen:
    //
    //     https://en.wikipedia.org/wiki/Half-power_point
    //     https://en.wikipedia.org/wiki/Decibel and https://www.rapidtables.com/electric/decibel.html
    //
    // Amplitude ratio 20*log10(Vin/Vout)
    //
    #define NUM_AMPLITUDE_STEPS        11 // 22
    #define _IOF_HEADSET_GAIN_DB_OFF    0
    #define _IOF_HEADSET_GAIN_DB_UNITY  4 // 11
    #define _IOF_HEADSET_GAIN_DB_MAX    (NUM_AMPLITUDE_STEPS-1)
    //
    #define SINE_480HZ_DIV_FACTOR       1000 // -60dB.    To protect the ears and the headset
    #define SINE_480HZ_SHL_BITS_FACTOR    10 // About -60dB. To protect the ears and the headset
    //
    #define AMPLITUDE_FIX16_UNITY_FACTORS_TABLE {  \
             0,                 /* OFF    */ \
         /* _HEADSET_GAIN_UNITY / 1000,          -80 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.00316,    -50 dB */ \
         /* _HEADSET_GAIN_UNITY /  100,          -40 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.032,      -30 dB */ \
         /* _HEADSET_GAIN_UNITY /   10,          -20 dB */ \
            _HEADSET_GAIN_UNITY *    0.316,   /* -10 dB */ \
            _HEADSET_GAIN_UNITY *    0.501,   /*  -6 dB (HALF 0.5 is -6.0206 dB) */ \
            _HEADSET_GAIN_UNITY *    0.708,   /*  -3 dB (1/2     is -3.0103 dB) */ \
         /* _HEADSET_GAIN_UNITY *    0.794,       -2 dB */ \
         /* _HEADSET_GAIN_UNITY *    0.891,       -1 dB */ \
            _HEADSET_GAIN_UNITY *    1,       /*   0 dB [_IOF_HEADSET_GAIN_DB_UNITY] */ \
         /* _HEADSET_GAIN_UNITY *    1.122,        1 dB */ \
         /* _HEADSET_GAIN_UNITY *    1.259,        2 dB */ \
            _HEADSET_GAIN_UNITY *    1.413,   /*   3 dB (2       is 3.0103 dB ) */ \
            _HEADSET_GAIN_UNITY *    1.995,   /*   6 dB (DOUBLE 2 is 6.0206 dB) */ \
            _HEADSET_GAIN_UNITY *    3.162,   /*  10 dB */ \
            _HEADSET_GAIN_UNITY *   10,       /*  20 dB */ \
            _HEADSET_GAIN_UNITY *   31.623,   /*  30 dB */ \
            _HEADSET_GAIN_UNITY *  100,       /*  40 dB */ \
         /* _HEADSET_GAIN_UNITY *  316.228,       50 dB */ \
         /* _HEADSET_GAIN_UNITY * 1000            60 dB */ \
    };
    #define IOF_HEADSET_GAIN_DB_DEFAULT _IOF_HEADSET_GAIN_DB_UNITY
    #define IOF_HEADSET_GAIN_DB_MAX     _IOF_HEADSET_GAIN_DB_MAX
    #define IOF_HEADSET_GAIN_DB_OFF     _IOF_HEADSET_GAIN_DB_OFF
    //
    #define HEADSET_GAIN_DEFAULT        _HEADSET_GAIN_UNITY
    #define HEADSET_GAIN_BITPOS_DEFAULT _HEADSET_GAIN_UNITY_BITPOS
    //
    #define HEADSET_GAIN_IS_OFF_VALUE_DEFAULT false;

    #define AMPLITUDE_DB_TABLE \
    { \
            -99, /* OFF */ \
            -10, \
             -6, \
             -3, \
              0, /* [_IOF_HEADSET_GAIN_DB_UNITY] */ \
              3, \
              6, \
             10, \
             20, \
             30, \
             40, \
    };

    typedef struct {
        unsigned        iof_gain;
        bool            headset_gain_is_off;
        headset_gain_t  gain;
        mic_data_t      amplitude_fix16_unity_factors_table [NUM_AMPLITUDE_STEPS];
        signed          amplitude_db_table                  [NUM_AMPLITUDE_STEPS];
    } headset_gain_context_t;

    void set_headset_gain (
            const do_headset_gain_e do_headset_gain,
            headset_gain_context_t      &ctx);

    #define I2C_ADDRESS_OF_DAC    0x4A // CS43L21, see page 5 of mic_array board manual
    #define CS41L21_DAC_ETH_RESET 0x03 // 4F0=DAC_RST_N Audio    CS43L21
                                       // 4F1=ETH_RST_N Ethernet LAN8710A
    //
    typedef enum i2c_reg_addr_dac_headphones_cs43l21_e {
        CS41L21_DAC_REG_01_CHIP_ID_AND_REVISION                      = 0x01, // USED
        CS41L21_DAC_REG_02_POWER_CONTROL_1                           = 0x02, // USED
        CS41L21_DAC_REG_03_SPEED_CONTROL                             = 0x03, // USED
        CS41L21_DAC_REG_04_INTERFACE_CONTROL                         = 0x04,
        CS41L21_DAC_REG_08_DAC_OUTPUT_CONTROL                        = 0x08,
        CS41L21_DAC_REG_09_DAC_CONTROL                               = 0x09,
        CS41L21_DAC_REG_10_MIXER_VOLUME_CONTROL_PCMA                 = 0x10, // USED
        CS41L21_DAC_REG_11_MIXER_VOLUME_CONTROL_PCMB                 = 0x11,
        CS41L21_DAC_REG_12_BEEP_FREQUENCY_AND_TIMING_CONFIGURATION   = 0x12,
        CS41L21_DAC_REG_13_BEEP_OFF_TIME_AND_VOLUME                  = 0x13,
        CS41L21_DAC_REG_14_BEEP_CONFIGURATION_AND_TONE_CONFIGURATION = 0x14,
        CS41L21_DAC_REG_15_TONE_CONTROL                              = 0x15,
        CS41L21_DAC_REG_16_VOLUME_CONTROL_AOUTA                      = 0x16,
        CS41L21_DAC_REG_17_VOLUME_CONTROL_AOUTB                      = 0x17,
        CS41L21_DAC_REG_18_PCM_CHANNEL_MIXER                         = 0x18,
        CS41L21_DAC_REG_19_LIMITER_THRESHOLD_SZC_DISABLE             = 0x19,
        CS41L21_DAC_REG_1A_LIMITER_RELEASE_RATE_REGISTER             = 0x1A,
        CS41L21_DAC_REG_1B_LIMITER_ATTACK_RATE_REGISTER              = 0x1B,
        CS41L21_DAC_REG_20_STATUS                                    = 0x20,
        CS41L21_DAC_REG_21_CHARGE_PUMP_FREQUENCY                     = 0x21
    } i2c_reg_addr_dac_headphones_cs43l21_e;

    #define I2C_ADDRESS_OF_PLL 0x4E // CS2100-CP, see page 5 of mic_array board manual
    //
    typedef enum i2c_reg_addr_pll_cs2100cp_e {
        CS2100CP_PLL_REG_01_DEVICE_ID_AND_REVISION   = 0x01,
        CS2100CP_PLL_REG_02_DEVICE_CONTROL           = 0x02,
        CS2100CP_PLL_REG_03_DEVICE_CONFIGURATION_1   = 0x03,
        CS2100CP_PLL_REG_05_GLOBAL_CONFIGURATION     = 0x05,
        CS2100CP_PLL_REG_06_RATIO                    = 0x06,
        CS2100CP_PLL_REG_07_RATIO                    = 0x07,
        CS2100CP_PLL_REG_08_RATIO                    = 0x08,
        CS2100CP_PLL_REG_09_RATIO                    = 0x09,
        CS2100CP_PLL_REG_16_FUNCTION_CONFIGURATION_1 = 0x16,
        CS2100CP_PLL_REG_17_FUNCTION_CONFIGURATION_2 = 0x17,
        CS2100CP_PLL_REG_1E_FUNCTION_CONFIGURATION_3 = 0x1E
    } i2c_reg_addr_pll_cs2100cp_e;

    [[distributable]]
    void i2s_task (
            server  i2s_callback_if i2s, // From i2s_master0
            client  i2c_master_if   i2c,
            port    p_rst_shared,
            chanend c_audio);

    // =============
    // COMMON IN-OUT
    // =============

    typedef struct mic_result_t { // TODO: needed?
        unsigned     num_samples;
        mic_sample_t mic_0_sample;
        time32_t     used_ticks;     // debug
        time32_t     max_used_ticks; // debug
    } mic_result_t;

    typedef struct menu_result_t {
        bool             headset_gain_updated;
        bool             headset_gain_is_off;
        headset_gain_t   headset_gain;
        bool             directive_mics_valid;
        directive_mics_t directive_mics;
        bool             clear_debug_vars;
    } menu_result_t;

    #define MAX_WAIT_TIME_TO_SEND_DATA_MS 40 // Dependent on SPEED_THROUGHPUT_TAG, I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND
                                             // and the load on the external I2C from logger_task, see DAC4_REG_FAST_WRITE_CMD_ABCD_NUM_BITS
                                             // Ver 0414 5Jan2022 (on button push and writing to display):
                                             //     33 -> [264]@ 48 kHz gives num_lost_overflowed 1 or 2
                                             //     34 -> [272]@ 48 kHz gives num_lost_overflowed always zero
                                             //     40 -> [320]@ 48 kHz good margin
    #define SAMPLES_BUF_DIM ((SAMPLING_FREQUENCY_HZ * MAX_WAIT_TIME_TO_SEND_DATA_MS) / (1000 * PROCESS_EVERY_N_SAMPLE))

    typedef struct {
         mic_sample_t arr[SAMPLES_BUF_DIM];
    } sch_ab_struct_with_array_t;

    #define SEND_VALUE_NEEDED_TICKS (6 * XS1_TIMER_MHZ) // Was 6 us for KnockCome. TODO: Not tested for "implementation F" (lkie 0413)
                                                        // 6 us means that calcs should get done in 20-6=14 us:
                                                        // SAMPLING_FREQUENCY_HZ 48000 -> f = (1/48000) = 20.833 us
                                                        // TESTED to be between 160 ns and 1.6 us, version 0235

    #define SPEED_THROUGHPUT_TAG // Not used in code, only made to search for. Depends on:
                                 // I2C_EXTERNAL_SPEED_K_BITS_PER_SECOND, SAMPLES_BUF_DIM

    // IMPLEMENTATION F
    //
    typedef unsigned num_mic_samples_t;

    typedef unsigned chan_mutual_packet_dim_e;
    // Using defines, since preprocessor will not resolve the tests below if not:
    //
    // This could have been done directly with [] if I used interface calls
    //
    // Modify in, should this list be changed:
    //   get_chan_mutual_packet_dim
    //   send_variant_protocol_then_shift_down
    //   receive_variant_protocol_send_clone_iff
    //
    #define NUM_000       0
    #define NUM_001       1
    #define NUM_002       2
    #define NUM_004       4
    #define NUM_008       8
    #define NUM_016      16
    #define NUM_032      32
    #define NUM_064      64
    #define NUM_MAX_MSG  90              // Max sent over channel when requested, phase II (N times first) (*)
    #define NUM_BUF      SAMPLES_BUF_DIM // Max buffered

    // (*)  Tested with 70, 92, 128 see version 0407, 0498, 0409. All work
    //      When DO_MEMFILL_TYPE == MEMFILL_TYPE_MEMCPY) in send_variant_protocol_then_shift_down
    //      with version v0423 then 90 seems to be a good value.

    #if (NUM_MAX_MSG >= NUM_BUF)
        #error Buffer must be the largest
    #endif

    //f ((SAMPLES_BUF_DIM - NUM_MAX_MSG) > (4 * NUM_MAX_MSG)) // v0800  320-90=230 < 4*90=360 ok
    #if ((SAMPLES_BUF_DIM - NUM_MAX_MSG) > (7 * NUM_MAX_MSG)) // v0801  640-90=550 < 7*90=560 ok
        #error Wrong strategy? Use ring buffer instead?
    #endif

    typedef struct { mic_sample_t samples [NUM_001];     } mic_samples_001_t;
    typedef struct { mic_sample_t samples [NUM_002];     } mic_samples_002_t;
    typedef struct { mic_sample_t samples [NUM_004];     } mic_samples_004_t;
    typedef struct { mic_sample_t samples [NUM_008];     } mic_samples_008_t;
    typedef struct { mic_sample_t samples [NUM_016];     } mic_samples_016_t;
    typedef struct { mic_sample_t samples [NUM_032];     } mic_samples_032_t;
    typedef struct { mic_sample_t samples [NUM_064];     } mic_samples_064_t;
    typedef struct { mic_sample_t samples [NUM_MAX_MSG]; } mic_samples_max_msg_t;
    typedef struct { mic_sample_t samples [NUM_BUF];     } mic_samples_buf_t;

    typedef struct {
        unsigned num_samples; // The number of valid samples
        unsigned num_lost_overflowed;
    } data_ch_nums_s;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_001_t mic_samples;
    } data_ch_sample_001_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_002_t mic_samples;
    } data_ch_sample_002_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_004_t mic_samples;
    } data_ch_sample_004_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_008_t mic_samples;
    } data_ch_sample_008_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_016_t mic_samples;
    } data_ch_sample_016_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_032_t mic_samples;
    } data_ch_sample_032_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_064_t mic_samples;
    } data_ch_sample_064_t;

    typedef struct {
        data_ch_nums_s        nums;
        mic_samples_max_msg_t mic_samples;
    } data_ch_sample_max_msg_t;

    typedef struct {
        data_ch_nums_s    nums;
        mic_samples_buf_t mic_samples;
    } data_ch_sample_buf_t;

    typedef struct { // sizeof these individually: 001: 12, 004: 24, 128: 520
        // No common fields here! Union has to cover all data!
        union {
            data_ch_sample_001_t     rec_001;
            data_ch_sample_002_t     rec_002;
            data_ch_sample_004_t     rec_004;
            data_ch_sample_008_t     rec_008;
            data_ch_sample_016_t     rec_016;
            data_ch_sample_032_t     rec_032;
            data_ch_sample_064_t     rec_064;
            data_ch_sample_max_msg_t rec_max_msg; // This and all the above
            data_ch_sample_buf_t     rec_buf;     // Not in data_ch_phase3_t
        } samples;
    } data_ch_all_samples_tx_and_buf_t;


    #define MIC_SAMPLE_VOID (0) // avoid this value when DEBUG_MIC_SAMPLES_INCREASING
    // This test would only be run for a short time, so wouldn't reach being wrapped around
    //
    #define COUNTING_MIC_SAMPLE_INIT       (MIC_SAMPLE_VOID + 1)          // 1
    #define COUNTING_MIC_SAMPLE_FIRST_SENT (COUNTING_MIC_SAMPLE_INIT + 1) // 2 (value implicitly used)

    typedef struct { // MAX_SIZEOF_NONBLOCKING_STREAMING_CHAN
        union {
            mic_sample_t mic_sample; // iff chan_mutual_packet_dim == NUM_000
            int32_t      is_void;    // iff chan_mutual_packet_dim >= NUM_001 (MIC_SAMPLE_VOID)
        } valid;
    } mic_sample_iff_t;


    // USAGE OF THESE TWO CHANNELS
    //   streaming chanend  sch_ab, // data_sch_phase1_t, data_ch_phase3_t
    //   chanend            ch_ba)  // data_ch_phase2_t
    //
    // "IMPLEMENTATION F" IN WORDS:
    // REPEAT --------------------------------------------------------------------
    //     WHEN task_a has something to send to task_b, either
    //        something new
    //        something next, or next again, ..
    //        something final
    //     PHASE1:       task_a will "ping" to task_b to say how much data is left
    //                   (max 2 words not to block on asynchronous buffered chan)
    //     PHASE2:       task_b eventually sends "ready" to task_a and atomically
    //                   goes to wait for response
    //                   (task_b may also just send "menu data" any time,
    //                   but this is not defined as a "PHASE")
    //     PHASE3:       task_a eventually receives the "ready" from task_b and
    //                   atomically sends the next batch of max size NUM_MAX_MSG
    //                   to task_b (used as synchronous unbuffered chan)
    //     CONTINUE_IFF: task_a will repeat (starting a new PHASE1) until it
    //                   has sent all SAMPLES_BUF_DIM data
    // FINALLY -------------------------------------------------------------------
    //     New mic samples that have arrived during this REPEAT
    //     are sent in a fast and final PHASE1-PHASE2-PHASE3
    // SAMPLES -------------------------------------------------------------------
    //     After this one and one sample is sent over inside PHASE1 "pings".
    //     PHASE2 is then delayed only if task_b is busy. During that time
    //     samples are buffered in task_a until it receives a "ready" from task_b
    //
    // DECOUPLE TASKS MESSAGE SEQUENCE DIAGRAM
    //   ==========                ==========
    //   | task_a |                | task_b |
    //   ==========                ==========
    //       | PHASE1                  |
    //       | asynch nonblock:        |
    //       | --> data_sch_phase1_t --> select case
    //       | work                    | do:
    //       |                 PHASE2  |
    //       <== data_ch_phase2_t <=====
    //       | select case             # by contract now wait for:
    //       # by contract now send:   #
    //       # PHASE3                  #
    //       ===> data_ch_phase3_t ====>
    //
    // =======
    // PHASE1
    // task_a nonblocking asynch message on chan for task_b, which will eventually pick it up in a select
    // task_a continues with any work
    // SPONTANEOUS START SEQUENCE. MAX 2 words on streaming chan not to block!
    //

    // "ping" with or without valid mic_sample
    typedef struct {
        chan_mutual_packet_dim_e chan_mutual_packet_dim;
        mic_sample_iff_t         iff; // Is valid mic_sample if above is NUM_000
        //
    } data_sch_phase1_t; // Observe MAX_SIZEOF_NONBLOCKING_STREAMING_CHAN, xassert done

    // =======
    // PHASE2
    // task_b chan to select in task_a, then immediate PHASE3
    // task_b waits for task_a in next line
    // RESPONSE BACK WHEN READY, any size of data on channel
    //
    typedef struct {
        chan_mutual_packet_dim_e chan_mutual_packet_dim;
        menu_result_t            menu_result;
    } data_ch_phase2_t;

    // =======
    // PHASE3
    // task_a chan to explicitly waiting task_b, ie. no blocking
    // NOW SEND MUCH DATA, but streaming chan will still not block
    //
    // Also send after data_ch_phase0_t in another direction
    //
    typedef struct { // sizeof these individually: 001: 12, 004: 24, 128: 520
        // No common fields here! Union has to cover all data!
        union {
            data_ch_sample_001_t     rec_001;
            data_ch_sample_002_t     rec_002;
            data_ch_sample_004_t     rec_004;
            data_ch_sample_008_t     rec_008;
            data_ch_sample_016_t     rec_016;
            data_ch_sample_032_t     rec_032;
            data_ch_sample_064_t     rec_064;
            data_ch_sample_max_msg_t rec_max_msg;
        } samples;
    } data_ch_phase3_t;

    typedef unsigned data_ch_init_synch_t; // data-less would have been ok, just for synch

    // ==== ONLY FOR THE DAC OUT ====
    //
    #define MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER (4)
    #define MASTER_CLOCK_FREQUENCY_HZ                24576000
    #define PDM_CLOCK_FREQUENCY_HZ                   (MASTER_CLOCK_FREQUENCY_HZ / (2 * MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER))
    #define DAC_OUTPUT_SAMPLE_FREQUENCY_HZ           (PDM_CLOCK_FREQUENCY_HZ / (32 * DECIMATION_FACTOR))
    #define MCLK_BCLK_RATIO                          (DECIMATION_FACTOR * MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER)
    //      ####_####_##### from these formulas:
    // mclk_bclk_ratio = (MASTER_CLOCK_FREQUENCY_HZ/DAC_OUTPUT_SAMPLE_FREQUENCY_HZ)/64
    // Reducing formula above:
    // D                 DECIMATION_FACTOR
    // A (4)             MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER
    // B 24576000        MASTER_CLOCK_FREQUENCY_HZ
    // C (B / (2 * A))   (MASTER_CLOCK_FREQUENCY_HZ / (2 * MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER))
    // E (C / (32 * D))  (PDM_CLOCK_FREQUENCY_HZ / (32 * DECIMATION_FACTOR))
    //
    // MCLK_BCLK_RATIO = (B/E)/64 = D * A =
    //                   DECIMATION_FACTOR * MASTER_PLL_24576KHZ_TO_PDM_CLOCK_DIVIDER



    // Only if DEBUG_MIC_SAMPLES_INCREASING!=0
    //
    void verify_mic_samples_delta ( // return newest_sample
            const unsigned     tag,
            const mic_sample_t delta,
            mic_sample_t       &oldest_sample,
            const mic_sample_t newest_sample,
            const signed       val);

    #if (DEBUG_MIC_SAMPLES_INCREASING == 0)
        #define VERIFY_MIC_SAMPLES_DELTA(t,d,o,n,v)
    #else
        #define VERIFY_MIC_SAMPLES_DELTA(t,d,o,n,v) do { verify_mic_samples_delta (t,d,o,n,v); } while (0)
    #endif

    void send_variant_protocol_then_shift_down (
            streaming chanend                sch_out,
            const chan_mutual_packet_dim_e   chan_mutual_packet_dim,
            data_ch_all_samples_tx_and_buf_t &data_ch_all_samples_tx_and_buf);

    void send_clone_iff (
            const bool              do_clone_out, // if true match receive_variant_protocol, from buffer_task_x
            const data_sch_phase1_t phase1_sch_mic_sample,
            chanend                 ch_out);

    void receive_variant_protocol_send_clone_iff (
            streaming chanend              sch_in,
            const chan_mutual_packet_dim_e chan_mutual_packet_dim,
            data_ch_phase3_t               &phase3_ch_mic_samples,
            const bool                     do_chan_out,
            const data_sch_phase1_t        phase1_sch_mic_sample,
            chanend                        ch_out);

    void receive_variant_protocol (
            chanend                        ch_in, //.. then data_ch_phase3_t here
            const chan_mutual_packet_dim_e chan_mutual_packet_dim,
            data_ch_phase3_t               &phase3_ch_mic_samples);

    [[combinable]]
    void headset_buffer_task (
            chanend ch_headset_in,
            chanend ch_headset_out);

    void mics_in_headset_out_task_a (
            streaming chanend ds_output_is_ptr_chan [NUM_4MIC_DECIMATORS], // [2/1] when MIC_ARRAY_NUM_MICS == 8/4
            mic_data_t        mics_data [NUM_DATA_X] [NUM_DATA_Y],         // [8/4] [192] when MIC_ARRAY_NUM_MICS == 8/4. Double-word aligned
            out port          p_scope_gray,
            out port          p_scope_1i_1m,
            chanend           ch_headset_out,
            streaming chanend sch_ab, // data_sch_phase1_t, data_ch_phase3_t
            chanend           ch_ba); // data_ch_phase2_t, data_ch_init_synch_t

    #ifndef DEBUG_MIC_SAMPLES_INCREASING
        #error include of _globals.h missing
    #endif

#else
    #error NESTED .h FILE. BUT IT NEEDS:
    /* TODO: make this list for all header files
    lib_mic_array included in makefile will resolve  local "mic_array_conf.h" to be seen
    #include <i2s.h>
    #include <i2c.h>
    #include "mic_array.h"
    #include "_globals.h"
    #include "maths_fix.h"
    */
#endif
