/*
 * port_tile0_task.xc
 *
 * None of the ports on the xCORE MICROPHONE ARRAY board's expansien header are on tile[1]
 * From tasks on tile[1] use this to excercise port(s)
 *
 *  Created on: 7. apr. 2022
 *      Author: oyvindteig
 */

#define INCLUDES
#ifdef INCLUDES
    #include <xs1.h>
    #include <platform.h> // tile
    #include <timer.h>    // delay_milliseconds(200), XS1_TIMER_HZ etc
    #include <stdint.h>   // uint8_t
    #include <stdio.h>    // printf
    #include <xscope.h>   // xScope logging
    #include <string.h>   // memcpy
    #include <xccompat.h> // REFERENCE_PARAM(my_app_ports_t, my_app_ports) -> my_app_ports_t &my_app_ports
    #include <iso646.h>   // not etc.
    #include <xassert.h>

    #include "_version.h" // First this..
    #include "_globals.h" // ..then this

    #include "port_tile0_task.h"

    #define DO_DEBUG_PRINT         0 // 0 has consequences if (USE_XSCOPE == 1) and amy timing at all needs to be met
    #define DO_DEBUG_PRINT_SPECIAL 0
    #define DO_WARNINGS            1
    #include "_print_macros_xc.h"
#endif

[[distributable]] // but [[combined]] here since on another tile
void port_tile0_task (
        server  port_commands_if if_port_commands[PORT_TILE0_TASK_NUM_CLIENTS],
        out port                  p_scope_1i_1m,
        out port                  p_scope_1d_1n)
{
    while (1) {
        select {
            case if_port_commands[int index_of_client].write_port (
                    const xs1_port_one_of_two_e xs1_port,
                    const port_e                value_bit0) : {

                // Page 63 of XMOS Programming Guide
                //     In general, when outputting to an n-bit port,
                //     the least significant n bits of the output value
                //     are driven on the pins and the rest are ignored.

                switch (xs1_port) {
                    case PORT_1I_1M: p_scope_1i_1m <: value_bit0; break;
                    case PORT_1D_1N: p_scope_1d_1n <: value_bit0; break;
                    default:         xassert (false);        break;
                };
            } break;
        }
    }
}
